@@ -15,7 +15,7 @@
  * with this program; if not, write to the Free Software Foundation, Inc.,
  * 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
  *
- * Copyright (C) 2008 - 2009 Dan Williams <dcbw@redhat.com> and Red Hat, Inc.
+ * Copyright (C) 2008 - 2011 Dan Williams <dcbw@redhat.com> and Red Hat, Inc.
  *
  **************************************************************************/
 
@@ -43,29 +43,33 @@
 #include "nm-openvpn.h"
 #include "../src/nm-openvpn-service.h"
 
-#define CLIENT_TAG "client"
-#define TLS_CLIENT_TAG "tls-client"
-#define DEV_TAG "dev "
-#define PROTO_TAG "proto "
-#define REMOTE_TAG "remote "
-#define CA_TAG "ca"
-#define CERT_TAG "cert"
-#define KEY_TAG "key"
-#define CIPHER_TAG "cipher"
-#define COMP_TAG "comp-lzo"
-#define IFCONFIG_TAG "ifconfig "
-#define SECRET_TAG "secret"
-#define AUTH_USER_PASS_TAG "auth-user-pass"
-#define TLS_AUTH_TAG "tls-auth"
 #define AUTH_TAG "auth "
-#define RENEG_SEC_TAG "reneg-sec"
-#define TLS_REMOTE_TAG "tls-remote"
-#define PORT_TAG "port"
-#define RPORT_TAG "rport"
+#define AUTH_USER_PASS_TAG "auth-user-pass"
+#define CA_TAG "ca "
+#define CERT_TAG "cert "
+#define CIPHER_TAG "cipher "
+#define CLIENT_TAG "client"
+#define COMP_TAG "comp-lzo"
+#define DEV_TAG "dev "
+#define FRAGMENT_TAG "fragment "
+#define IFCONFIG_TAG "ifconfig "
+#define KEY_TAG "key "
 #define MSSFIX_TAG "mssfix"
-#define TUNMTU_TAG "tun-mtu"
-#define FRAGMENT_TAG "fragment"
-#define PKCS12_TAG "pkcs12"
+#define PKCS12_TAG "pkcs12 "
+#define PORT_TAG "port "
+#define PROTO_TAG "proto "
+#define HTTP_PROXY_TAG "http-proxy "
+#define HTTP_PROXY_RETRY_TAG "http-proxy-retry"
+#define SOCKS_PROXY_TAG "socks-proxy "
+#define SOCKS_PROXY_RETRY_TAG "socks-proxy-retry"
+#define REMOTE_TAG "remote "
+#define RENEG_SEC_TAG "reneg-sec "
+#define RPORT_TAG "rport "
+#define SECRET_TAG "secret "
+#define TLS_AUTH_TAG "tls-auth "
+#define TLS_CLIENT_TAG "tls-client"
+#define TLS_REMOTE_TAG "tls-remote "
+#define TUNMTU_TAG "tun-mtu "
 
 
 static char *
@@ -197,6 +201,58 @@ parse_port (const char *str, const char *line)
 	return NULL;
 }
 
+static gboolean
+parse_http_proxy_auth (const char *path,
+                       const char *file,
+                       char **out_user,
+                       char **out_pass)
+{
+	char *contents = NULL, *abspath = NULL, *tmp;
+	GError *error = NULL;
+	char **lines, **iter;
+
+	g_return_val_if_fail (out_user != NULL, FALSE);
+	g_return_val_if_fail (out_pass != NULL, FALSE);
+
+	if (!file || !strcmp (file, "stdin") || !strcmp (file, "auto") || !strcmp (file, "'auto'"))
+		return TRUE;
+
+	if (!g_path_is_absolute (file)) {
+		tmp = g_path_get_dirname (path);
+		abspath = g_build_path ("/", tmp, file, NULL);
+		g_free (tmp);
+	} else
+		abspath = g_strdup (file);
+
+	/* Grab user/pass from authfile */
+	if (!g_file_get_contents (abspath, &contents, NULL, &error)) {
+		g_warning ("%s: unable to read HTTP proxy authfile '%s': (%d) %s",
+		           __func__, abspath, error ? error->code : -1,
+		           error && error->message ? error->message : "(unknown)");
+		g_clear_error (&error);
+		g_free (abspath);
+		return FALSE;
+	}
+
+	lines = g_strsplit_set (contents, "\n\r", 0);
+	for (iter = lines; iter && *iter; iter++) {
+		if (!strlen (*iter))
+			continue;
+		if (!*out_user)
+			*out_user = g_strdup (g_strstrip (*iter));
+		else if (!*out_pass) {
+			*out_pass = g_strdup (g_strstrip (*iter));
+			break;
+		}
+	}
+	if (lines)
+		g_strfreev (lines);
+	g_free (contents);
+	g_free (abspath);
+
+	return *out_user && *out_pass;
+}
+
 NMConnection *
 do_import (const char *path, char **lines, GError **error)
 {
@@ -209,7 +265,8 @@ do_import (const char *path, char **lines, GError **error)
 	gboolean have_pass = FALSE, have_sk = FALSE;
 	const char *ctype = NULL;
 	char *basename;
-	char *default_path;
+	char *default_path, *tmp, *tmp2;
+	gboolean http_proxy = FALSE, socks_proxy = FALSE, proxy_set = FALSE;
 
 	connection = nm_connection_new ();
 	s_con = NM_SETTING_CONNECTION (nm_setting_connection_new ());
@@ -221,7 +278,15 @@ do_import (const char *path, char **lines, GError **error)
 	
 	/* Get the default path for ca, cert, key file, these files maybe
 	 * in same path with the configuration file */
-	default_path = g_path_get_dirname (path);
+	if (g_path_is_absolute (path))
+		default_path = g_path_get_dirname (path);
+	else {
+		tmp = g_get_current_dir ();
+		tmp2 = g_path_get_dirname (path);
+		default_path = g_build_filename (tmp, tmp2, NULL);
+		g_free (tmp);
+		g_free (tmp2);
+	}
 
 	basename = g_path_get_basename (path);
 	last_dot = strrchr (basename, '.');
@@ -231,7 +296,7 @@ do_import (const char *path, char **lines, GError **error)
 	g_free (basename);
 
 	for (line = lines; *line; line++) {
-		char *comment, **items, *leftover = NULL;
+		char *comment, **items = NULL, *leftover = NULL;
 
 		if ((comment = strchr (*line, '#')))
 			*comment = '\0';
@@ -282,7 +347,7 @@ do_import (const char *path, char **lines, GError **error)
 				errno = 0;
 				secs = strtol (items[0], NULL, 10);
 				if ((errno == 0) && (secs >= 0) && (secs < 0xffff)) {
-					char *tmp = g_strdup_printf ("%d", (guint32) secs);
+					tmp = g_strdup_printf ("%d", (guint32) secs);
 					nm_setting_vpn_add_data_item (s_vpn, NM_OPENVPN_KEY_TUNNEL_MTU, tmp);
 					g_free (tmp);
 				} else
@@ -303,7 +368,7 @@ do_import (const char *path, char **lines, GError **error)
 				errno = 0;
 				secs = strtol (items[0], NULL, 10);
 				if ((errno == 0) && (secs >= 0) && (secs < 0xffff)) {
-					char *tmp = g_strdup_printf ("%d", (guint32) secs);
+					tmp = g_strdup_printf ("%d", (guint32) secs);
 					nm_setting_vpn_add_data_item (s_vpn, NM_OPENVPN_KEY_FRAGMENT_SIZE, tmp);
 					g_free (tmp);
 				} else
@@ -328,8 +393,8 @@ do_import (const char *path, char **lines, GError **error)
 
 				errno = 0;
 				secs = strtol (items[0], NULL, 10);
-				if ((errno == 0) && (secs >= 0) && (secs < 604800)) {
-					char *tmp = g_strdup_printf ("%d", (guint32) secs);
+				if ((errno == 0) && (secs >= 0) && (secs <= 604800)) {
+					tmp = g_strdup_printf ("%d", (guint32) secs);
 					nm_setting_vpn_add_data_item (s_vpn, NM_OPENVPN_KEY_RENEG_SECONDS, tmp);
 					g_free (tmp);
 				} else
@@ -339,6 +404,72 @@ do_import (const char *path, char **lines, GError **error)
 			continue;
 		}
 
+		if (   !strncmp (*line, HTTP_PROXY_RETRY_TAG, strlen (HTTP_PROXY_RETRY_TAG))
+		    || !strncmp (*line, SOCKS_PROXY_RETRY_TAG, strlen (SOCKS_PROXY_RETRY_TAG))) {
+			nm_setting_vpn_add_data_item (s_vpn,
+			                              g_strdup (NM_OPENVPN_KEY_PROXY_RETRY),
+			                              g_strdup ("yes"));
+			continue;
+		}
+
+		http_proxy = g_str_has_prefix (*line, HTTP_PROXY_TAG);
+		socks_proxy = g_str_has_prefix (*line, SOCKS_PROXY_TAG);
+		if ((http_proxy || socks_proxy) && !proxy_set) {
+			gboolean success = FALSE;
+			const char *proxy_type = NULL;
+
+			if (http_proxy) {
+				items = get_args (*line + strlen (HTTP_PROXY_TAG));
+				proxy_type = "http";
+			} else if (socks_proxy) {
+				items = get_args (*line + strlen (SOCKS_PROXY_TAG));
+				proxy_type = "socks";
+			}
+			if (!items)
+				continue;
+
+			if (g_strv_length (items) >= 2) {
+				glong port;
+				char *s_port = NULL;
+				char *user = NULL, *pass = NULL;
+
+				success = TRUE;
+				if (http_proxy && g_strv_length (items) >= 3)
+					success = parse_http_proxy_auth (path, items[2], &user, &pass);
+
+				if (success) {
+					success = FALSE;
+					errno = 0;
+					port = strtol (items[1], NULL, 10);
+					if ((errno == 0) && (port > 0) && (port < 65536)) {
+						s_port = g_strdup_printf ("%d", (guint32) port);
+						success = TRUE;
+					}
+				}
+
+				if (success && proxy_type) {
+					nm_setting_vpn_add_data_item (s_vpn, NM_OPENVPN_KEY_PROXY_TYPE, proxy_type);
+
+					nm_setting_vpn_add_data_item (s_vpn, NM_OPENVPN_KEY_PROXY_SERVER, items[0]);
+					nm_setting_vpn_add_data_item (s_vpn, NM_OPENVPN_KEY_PROXY_PORT, s_port);
+					if (user)
+						nm_setting_vpn_add_data_item (s_vpn, NM_OPENVPN_KEY_HTTP_PROXY_USERNAME, user);
+					if (pass)
+						nm_setting_vpn_add_secret (s_vpn, NM_OPENVPN_KEY_HTTP_PROXY_PASSWORD, pass);
+					proxy_set = TRUE;
+				}
+				g_free (s_port);
+				g_free (user);
+				g_free (pass);
+			}
+
+			if (!success)
+				g_warning ("%s: invalid proxy option '%s'", __func__, *line);
+
+			g_strfreev (items);
+			continue;
+		}
+
 		if (!strncmp (*line, REMOTE_TAG, strlen (REMOTE_TAG))) {
 			items = get_args (*line + strlen (REMOTE_TAG));
 			if (!items)
@@ -349,8 +480,6 @@ do_import (const char *path, char **lines, GError **error)
 				have_remote = TRUE;
 
 				if (g_strv_length (items) >= 2) {
-					char *tmp;
-
 					tmp = parse_port (items[1], *line);
 					if (tmp) {
 						nm_setting_vpn_add_data_item (s_vpn, NM_OPENVPN_KEY_PORT, tmp);
@@ -367,8 +496,6 @@ do_import (const char *path, char **lines, GError **error)
 
 		if (   !strncmp (*line, PORT_TAG, strlen (PORT_TAG))
 		    || !strncmp (*line, RPORT_TAG, strlen (RPORT_TAG))) {
-			char *tmp;
-
 			/* Port specified in 'remote' always takes precedence */
 			if (nm_setting_vpn_get_data_item (s_vpn, NM_OPENVPN_KEY_PORT))
 				continue;
@@ -409,6 +536,7 @@ do_import (const char *path, char **lines, GError **error)
 			                  NM_OPENVPN_KEY_STATIC_KEY_DIRECTION,
 			                  leftover,
 			                  s_vpn);
+			have_sk = TRUE;
 			continue;
 		}
 
@@ -475,9 +603,6 @@ do_import (const char *path, char **lines, GError **error)
 		}
 	}
 
-	if (nm_setting_vpn_get_data_item (s_vpn, NM_OPENVPN_KEY_STATIC_KEY))
-		have_sk = TRUE;
-
 	if (!have_client && !have_sk) {
 		g_set_error (error,
 		             OPENVPN_PLUGIN_UI_ERROR,
@@ -558,6 +683,10 @@ do_export (const char *path, NMConnection *connection, GError **error)
 	gboolean use_lzo = FALSE;
 	gboolean reneg_exists = FALSE;
 	guint32 reneg = 0;
+	const char *proxy_type = NULL;
+	const char *proxy_server = NULL;
+	const char *proxy_port = NULL;
+	const char *proxy_retry = NULL;
 
 	s_con = NM_SETTING_CONNECTION (nm_connection_get_setting (connection, NM_TYPE_SETTING_CONNECTION));
 	g_assert (s_con);
@@ -737,6 +866,27 @@ do_export (const char *path, NMConnection *connection, GError **error)
 		}
 	}
 
+	/* Proxy stuff */
+	proxy_type = nm_setting_vpn_get_data_item (s_vpn, NM_OPENVPN_KEY_PROXY_TYPE);
+	if (proxy_type && strlen (proxy_type)) {
+		proxy_server = nm_setting_vpn_get_data_item (s_vpn, NM_OPENVPN_KEY_PROXY_SERVER);
+		proxy_port = nm_setting_vpn_get_data_item (s_vpn, NM_OPENVPN_KEY_PROXY_PORT);
+		proxy_retry = nm_setting_vpn_get_data_item (s_vpn, NM_OPENVPN_KEY_PROXY_RETRY);
+		if (!strcmp (proxy_type, "http") && proxy_server && proxy_port) {
+			if (!proxy_port)
+				proxy_port = "8080";
+			fprintf (f, "http-proxy %s %s\n", proxy_server, proxy_port);
+			if (proxy_retry && !strcmp (proxy_retry, "yes"))
+				fprintf (f, "http-proxy-retry\n");
+		} else if (!strcmp (proxy_type, "socks") && proxy_server && proxy_port) {
+			if (!proxy_port)
+				proxy_port = "1080";
+			fprintf (f, "socks-proxy %s %s\n", proxy_server, proxy_port);
+			if (proxy_retry && !strcmp (proxy_retry, "yes"))
+				fprintf (f, "socks-proxy-retry\n");
+		}
+	}
+
 	/* Add hard-coded stuff */
 	fprintf (f,
 	         "nobind\n"
